//   Copyright 2022 Esri
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.

package com.example.app;

import javafx.application.Application;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.geometry.Insets;
import javafx.geometry.Pos;
import javafx.scene.Scene;
import javafx.scene.control.Alert;
import javafx.scene.control.ComboBox;
import javafx.scene.control.SingleSelectionModel;
import javafx.scene.layout.Background;
import javafx.scene.layout.BackgroundFill;
import javafx.scene.layout.CornerRadii;
import javafx.scene.layout.StackPane;
import javafx.scene.layout.VBox;
import javafx.stage.Stage;
import javafx.scene.paint.Paint;
import javafx.scene.paint.Color;
import javafx.scene.control.Label;

import com.esri.arcgisruntime.ArcGISRuntimeEnvironment;
import com.esri.arcgisruntime.concurrent.ListenableFuture;
import com.esri.arcgisruntime.data.Feature;
import com.esri.arcgisruntime.data.FeatureQueryResult;
import com.esri.arcgisruntime.data.FeatureTable;
import com.esri.arcgisruntime.data.QueryParameters;
import com.esri.arcgisruntime.data.ServiceFeatureTable;
import com.esri.arcgisruntime.geometry.Envelope;
import com.esri.arcgisruntime.layers.FeatureLayer;
import com.esri.arcgisruntime.layers.Layer;
import com.esri.arcgisruntime.loadable.LoadStatus;
import com.esri.arcgisruntime.mapping.ArcGISMap;
import com.esri.arcgisruntime.mapping.BasemapStyle;
import com.esri.arcgisruntime.mapping.Viewpoint;
import com.esri.arcgisruntime.mapping.view.MapView;

public class App extends Application {

  private ComboBox<String> parcelsComboBox;

  private MapView mapView;

  public static void main(String[] args) {
    Application.launch(args);
  }

  @Override
  public void start(Stage stage) {

    // set the title and size of the stage and show it
    stage.setTitle("Query a feature layer (SQL)");

    stage.setWidth(800);
    stage.setHeight(700);
    stage.show();

    // create a JavaFX scene with a stack pane as the root node, and set it on the stage
    StackPane stackPane = new StackPane();
    Scene scene = new Scene(stackPane);
    stage.setScene(scene);

    // Note: it is not best practice to store API keys in source code.
    // The API key is referenced here for the convenience of this tutorial.
    String yourApiKey = "YOUR_API_KEY";
    ArcGISRuntimeEnvironment.setApiKey(yourApiKey);

    // create a map view to display the map, and add it to the stack pane
    mapView = new MapView();
    stackPane.getChildren().add(mapView);

    ArcGISMap map = new ArcGISMap(BasemapStyle.ARCGIS_TOPOGRAPHIC);
    mapView.setMap(map);

    mapView.setViewpoint(new Viewpoint(34.02700, -118.80543, 72000));

    // create a parcels combo box and label, and add to vertical controls box
    setupUI(stackPane);

    // create parcels feature layer
    createFeatureLayer(map);

  }

  // stops and releases all resources used in application
  @Override
  public void stop() {
    if (mapView != null) {
      mapView.dispose();
    }
  }

   // Creates a UI with a combo box.
  private void setupUI(StackPane stackPane) {

    // create a label
    Label parcelsComboBoxLabel = new Label("WHERE EXPRESSION: ");
    parcelsComboBoxLabel.setTextFill(Color.WHITE);

    // create the list of expressions for the combo box and then create the combo box
    ObservableList<String> parcelsComboBoxList = FXCollections.observableArrayList("Select an expression",
        "UseType = 'Government'", "UseType = 'Residential'", "UseType = 'Irrigated Farm'", "TaxRateArea = 10853",
        "TaxRateArea = 10860", "Roll_LandValue > 1000000", "Roll_LandValue < 1000000");
    parcelsComboBox = new ComboBox<>(parcelsComboBoxList);
    parcelsComboBox.setMaxWidth(Double.MAX_VALUE);

    // set the default combo box value
    parcelsComboBox.getSelectionModel().select(0);

    // set up the control panel UI
    VBox controlsVBox = new VBox(6);
    controlsVBox.setBackground(new Background(new BackgroundFill(Paint.valueOf("rgba(0, 0, 0, 0.3)"),
        CornerRadii.EMPTY, Insets.EMPTY)));
    controlsVBox.setPadding(new Insets(10.0));
    controlsVBox.setMaxSize(210, 50);
    controlsVBox.setVisible(true);

    // add the label and combo box to the control box
    controlsVBox.getChildren().addAll(parcelsComboBoxLabel, parcelsComboBox);

    // add the control box to the stack pane and set its alignment and margins
    stackPane.getChildren().add(controlsVBox);
    StackPane.setAlignment(controlsVBox, Pos.TOP_LEFT);
    StackPane.setMargin(controlsVBox, new Insets(10,0,0,10));

  }

  // Query the feature layer, providing a Where expression and the current extent.
  // Then select the features returned by the query.
  private void queryFeatureLayer(String featureLayerId, String whereExpression, Envelope queryExtent) {

    try {
      FeatureLayer featureLayerToQuery = null;

      // get the layer based on its Id
      for (Layer layer : mapView.getMap().getOperationalLayers()) {
        if (layer.getId().equals(featureLayerId)) {
          featureLayerToQuery = (FeatureLayer) layer;
          break;
        }
      }

      // check if feature layer retrieved based on layer Id
      if (featureLayerToQuery == null){
          String msg = "Specified Id did not match any feature layer in this map";
          new Alert(Alert.AlertType.ERROR, msg).show();
          return;
      }

      // get the feature table from the feature layer
      FeatureTable featureTableToQuery = featureLayerToQuery.getFeatureTable();

      // clear any previous selections
      featureLayerToQuery.clearSelection();

      // create a query for the state that was entered
      QueryParameters query = new QueryParameters();
      query.setWhereClause(whereExpression);
      query.setReturnGeometry(true);
      query.setGeometry(queryExtent);

      // call query features
      ListenableFuture<FeatureQueryResult> future = featureTableToQuery.queryFeaturesAsync(query);

      // create an effectively final variable for access from add done listener
      FeatureLayer finalFeatureLayerToQuery = featureLayerToQuery;

      // add done listener to fire when the query returns
      future.addDoneListener(() -> {
        try {
          // check if there are some results
          FeatureQueryResult featureQueryResult = future.get();
          if (featureQueryResult.iterator().hasNext()) {
            for (Feature feature : featureQueryResult) {
              finalFeatureLayerToQuery.selectFeature(feature);
            }
          } else {
            String msg = "No parcels found in the current extent, using Where expression: " + whereExpression + ".";
            new Alert(Alert.AlertType.INFORMATION, msg).show();
          }
        }
        catch (Exception e) {
          String msg = "Feature search failed for: " + whereExpression + ". Error: " + e.getMessage();
          new Alert(Alert.AlertType.ERROR, msg).show();
          e.printStackTrace();
        }
      });

    } catch (Exception e) {
        // on any error, display the stack trace
        e.printStackTrace();
    }

  }

   // Create and load the parcels feature layer from the feature service. When the layer is loaded, query it based on
   // current extent and current selection in parcelsComboBox
  private void createFeatureLayer(ArcGISMap map){

    FeatureTable serviceFeatureTable =
        new ServiceFeatureTable("https://services3.arcgis.com/GVgbJbqm8hXASVYi/arcgis/rest/services/LA_County_Parcels/FeatureServer/0");

    FeatureLayer layer = new FeatureLayer(serviceFeatureTable);

    // give the layer an ID, so we can easily find it later
    layer.setId("Parcels");

    // Load the layer and add a done loading listener, which runs only when the layer is completely loaded.
    layer.loadAsync();
    layer.addDoneLoadingListener( () -> {
      if (layer.getLoadStatus() == LoadStatus.LOADED) {
        // get the selected item property from the combo box, and add a listener that runs
        // when the property value is changed by the user.
        SingleSelectionModel<String> selectionModel = parcelsComboBox.getSelectionModel();
        selectionModel.selectedItemProperty().addListener(observable -> {
          if (selectionModel.getSelectedIndex() == 0) {
            layer.clearSelection();
            return;
          }
          String currentParcelsChoice = selectionModel.getSelectedItem();

          Envelope currentExtent =
              (Envelope) (mapView.getCurrentViewpoint(Viewpoint.Type.BOUNDING_GEOMETRY).getTargetGeometry());

          queryFeatureLayer(layer.getId(), currentParcelsChoice, currentExtent);
        });
      }
    });

    //  add the layer to the map
    map.getOperationalLayers().add(layer);

    mapView.getSelectionProperties().setColor(Color.YELLOW);

  }

}

